/*
  parseQual.c

  $Id: parseQual.c,v 1.1.1.1 2005/08/04 14:32:29 mcschatz Exp $
*/

#include "getCoverage.h"

FILE * f_qual = NULL;
int _allowMissing = 1;
int _doGapQV = 0;

void setDoGapQV(doGapQV)
{
  _doGapQV = doGapQV;
}

void closeQualFile()
{
  if (f_qual) { fclose(f_qual); }
}

void setQualFile(FILE * fqual)
{
  f_qual = fqual;
}

/*
  void parseQual()

  parse prefix.qual file and check that each sequence exists.
*/

void parseQual(Offset *index,
               int numOfQVSeqs,
               int numOfSeqs,
               struct CONTIG_SEQUENCE **bcseq,
               char * contig_id,
               char allowMissing,
               int doGapQV)
{
  int i;
  _allowMissing = allowMissing;
  
  // Just make sure each read has qualities in the index
  for(i = 0; i < numOfSeqs; i++)
  {
    int indexpos = binsearchOffset(bcseq[i]->id,index,numOfQVSeqs);

    if ((indexpos == -1) && (allowMissing == -1))
    {
      printError("parseQual", DEBUG_ERROR,
                 "ERROR: %s referenced in contig file but not found in qual file!",
                 bcseq[i]->id);
             
      exit(1); // No qualities found
    }

    if (indexpos == -1)
    {
      bcseq[i]->qvseqlen = -1;
      bcseq[i]->qvoffset = -1;
    }
    else
    {
      bcseq[i]->qvseqlen = index[indexpos].qvseqlen;
      bcseq[i]->qvoffset = index[indexpos].offset;
    }
  }
} // end of parseQual()






void loadQual(struct CONTIG_SEQUENCE *bcseq, char * contig_id)
{
  int len; 
  int clr1, clr2; 
  int bcseqpos,seqpos,gap;              

  char c;

  qvdata_t MINQV = 4;  
  qvdata_t swap;

  // gapped quality values sequence
  qvdata_t * gqv = pool_malloc(bcseq->len * sizeof(qvdata_t));  

  int baseslen = bcseq->len - bcseq->numOfGaps;
  off_t offset = bcseq->qvoffset;

  clr1 = bcseq->seq_l;
  if (bcseq->seq_r < clr1) { clr1 = bcseq->seq_r; }

  clr1--; // counts the number of bases that can't be used

  if ((offset < 0) ||
      (bcseq->seq_l == 0 && bcseq->seq_r == 0) ||
      (bcseq->qvseqlen - clr1 < baseslen))
  {
    if (offset < 0)
    {
      logMsg(DEBUG_HIGH, "parseQual",
             "Qual record for %s not found, using dummy values", bcseq->id);
    }
    else if (bcseq->qvseqlen - clr1 < baseslen)
    {
      printWarning("parseQual",
                   "Valid QV sequence len (%d) < bases len (%d) for %s, using dummy values",
                   bcseq->qvseqlen-clr1, 
                   baseslen,
                   bcseq->id);
    }
    else
    {
      printWarning("parseQual",
                   "Null CLR range for %s, using dummy values",
                   bcseq->id);
    }

    memset(gqv, (_allowMissing==-1)?0:_allowMissing, bcseq->len);  // qvdata_t
  }
  else // actually load the data
  {
    qvdata_t * qv = (qvdata_t *) pool_malloc(bcseq->qvseqlen * sizeof(qvdata_t));
    char digitsRead = 0;

    // Position to start of sequence name 
    fseeko(f_qual, offset, SEEK_SET);

    // verify the name of the read in case of index corruption
    seqpos = 0;

    while ((c=fgetc(f_qual)) != EOF && !isspace(c))
    {
      if (bcseq->id[seqpos] != c)
      {
        printError("parseQual", DEBUG_ERROR,
                   "Qual Index corrupted for %s, reported to be at offset:%lld",
                   bcseq->id, offset);
        exit(1); // qual index corrupted
      }

      seqpos++;
    }

    // advance past any comments
    while (c != EOF && c != '\n') { c = fgetc(f_qual); }
    ungetc(c, f_qual);

    seqpos = 0;
    len = bcseq->qvseqlen; // Don't trust this value

    // now load the actual qual data
    do
    {
      c = fgetc(f_qual); 
      if (isdigit(c))
      {
        int num = c - '0';

        if (digitsRead)
        {
          // in 45 case => 4 * 10 + 5 == 45
          // in 00 case => 0 * 10 + 0 == 0

          if (seqpos < len)
          {
            qv[seqpos] *= 10;
            qv[seqpos] += num;  
          }
        }
        else
        {
          if (seqpos < len)
          {
            // just set qv = num
            qv[seqpos] = num;
          }

          digitsRead = 1;
        }
      }
      else // EOF or '>' is okay
      {
        if (digitsRead) 
        { 
          digitsRead = 0;

          if (seqpos < len)
          {
            if (qv[seqpos] == 0) { qv[seqpos] = MINQV; }
          }

          seqpos++;
        }
      }
    } while (c != EOF && c != '>');

    len = seqpos; // len is the exact count which should match qvseqlen

    if (len != bcseq->qvseqlen)
    {
      printError("loadQual", DEBUG_ERROR,
                 "Qual idx file corruption detected.\n%s is supposed to have %d quals, but really has %d",
                 bcseq->id, bcseq->qvseqlen, len);
      exit(1); // idx corrupted
    }

    // reverse the sequence if rc = 1
    if(bcseq->rc)
    { 
      int last;
      // reverse qv
      for(seqpos=0, last=len-1; seqpos < last; seqpos++, last--)
      {
        swap = qv[seqpos];
        qv[seqpos] = qv[last];
        qv[last] = swap;
      }
    }

    // computing clear range limits (taking into account rc)
    clr1 = (bcseq->rc)? len - bcseq->seq_l + 1 : bcseq->seq_l;
    clr2 = (bcseq->rc)? len - bcseq->seq_r + 1 : bcseq->seq_r;

    if (clr2 < clr1)
    {
      printWarning("parseQual",
                   "End of clear range %d before beginning %d "\
                   "in contig %s, seq %s, reversing.", 
                   clr2, clr1, contig_id, bcseq->id);

      swap = clr1;
      clr1 = clr2;
      clr2 = swap;
    }

    clr1--; // Convert to 0-based

    // restricting qv to the current clear range and
    
    // fill the intervals between gaps (if any) with base calls
    // at the end (where the gap is) place gapqv
    // NOTE that gapInd values are derived from base call sequence
    // already restricted to its clear range

    for(bcseqpos = 0, seqpos = clr1, gap = 0; 
        gap < bcseq->numOfGaps; 
        gap++)
    { 
      // Copy the qvseq until there is a gap
      for(; 
          seqpos < clr1 + bcseq->gapInd[gap]; 
          bcseqpos++,seqpos++)
      {
        if (seqpos >= len)
        {
          printError("parseQual", DEBUG_ERROR,
                     "INTERNAL ERROR: bcseq > qvs %s %d-%dg=%d, {%d,%d}, qvlen:%d = %d\n",
                     bcseq->id, bcseq->len, bcseq->numOfGaps,
                     bcseq->len - bcseq->numOfGaps,
                     bcseq->seq_l, bcseq->seq_r,
                     len, len - bcseq->seq_l);
          exit(1); // internal error
        }

        gqv[bcseqpos] = qv[seqpos];
      }

      // Assign the gap a qv
      if (_doGapQV)
      {
        int gapqv = 20;
        int left  = (seqpos > 0)   ? qv[seqpos-1] : -1;
        int right = (seqpos < len) ? qv[seqpos]   : -1;

        if (left != -1 || right != -1)
        {
          gapqv = (left <= right) 
                  ? ((left != -1) ? left : right)
                  : ((right != -1) ? right : left);
        }

        // assign the gap qv as the minimum of the flanking qvs
        gqv[bcseqpos] = gapqv;
      }
      else
      {
        gqv[bcseqpos] = 0;
      }

      bcseqpos++;
    }

    // Copy the sequence after the last gap
    // This also handles the case that there are no gaps or clear range
    // adjustments
    for( ; bcseqpos < bcseq->len; bcseqpos++,seqpos++)
    {
      if (seqpos >= len)
      {
        printError("parseQual", DEBUG_ERROR,
                   "INTERNAL ERROR: bcseq > qvs %s %d-%dg=%d, {%d,%d}, qvlen:%d = %d\n",
                   bcseq->id, bcseq->len, bcseq->numOfGaps,
                   bcseq->len - bcseq->numOfGaps,
                   bcseq->seq_l, bcseq->seq_r,
                   len, len - bcseq->seq_l);
        exit(1); // internal error
      }

      gqv[bcseqpos] = qv[seqpos];
    }
  }

  bcseq->qvseq = gqv;
}
