#!/usr/bin/perl
# $Id: findSNPs.pl,v 1.1.1.1 2005/08/04 14:34:35 mcschatz Exp $
#
# findSNPs - Obtain a list of SNPs and indels from the nucmer alignment
#            of a query sequence to a reference sequence.  
#
# Written by -  Martin Shumway
#
my $HELPTEXT = qq~
  Obtain a list of SNPs and indels from the nucmer alignment
  of a query sequence to a reference sequence.  The alignment as 
  output from show-aligns is emitted as a series of SNP and indel
  records in a script-ready format.  These .aligns files can be
  merged together to present the SNP alignments of a query 
  organism to a reference organism.
  
  
    findSNPs  alignfile  [options] 
 
      alignfile    Filename of the .align file from show-aligns

      options: 
        --flank N   Add N bases on either side of motif (default flank 0)
        --[no]path  Display full pathname of operand (default --nopath)
        --chain N  Specify the deltas from chain N (0-based) only

  Output is in CSV format.
    col 1 - The filename of the reference sequence (from .aligns file)
    col 2 - The fasta header of the reference sequence (from nucmer)
    col 3 - The filename of the query sequence (from .aligns file)
    col 4 - The fasta header of the query sequence (from nucmer)
    col 5 - The chain id (if there are multiple alignments in each file)
            (The first chain is designated '0').
    col 6 - Start of alignment in reference (1-based)
    col 7 - End of alignment in reference (inclusive)
    col 8 - Start of alignment in query (1-based, forward or reverse)
    col 9 - End of alignment in query (inclusive)
    col 10- The strand of the query sequence (R=reverse strand) 
    col 11- Ungapped index from 1 of feature in reference sequence
            If indel, motif starts after index 
    col 12- Ungapped index from 1 of feature in query sequence
            If indel, motif starts after index regardless of strand
    col 13- Ambiguity code of SNP, or sequence of indel motif
    col 14- Whether transition (1) or transversion (0), or (2) if neither
    col 15- Polymorphism type (S=SNP, I=insertion in query sequence,
            D=deletion from query sequence, N=none (perfect alignment))
    col 16- Flanking bases around SNP or indel in reference sequence 
    col 17- Flanking bases around SNP or indel in query sequence 

  See Also:
    show-aligns - Part of MUMmer 2.13 package
~;
#
# (C) Copyright 2003 - 2004 The Institute for Genomic Research (TIGR)
#     All rights reserved.  Written for CS452 Biological Databases
#     (Johns Hopkins University) Spring 2003.
#
#

# =============================== Pragmas and imports ======================
use strict;
use IO::File;
use File::Basename;
use TIGR::Foundation;

# Normally used only in testing
# use warnings;

# ================================ Globals ====================================

# Program stuff
my $MY_APP = basename($0);
my $MY_LOG = "$MY_APP.log";
my $MY_VERSION = " Version 1.42 (Build " . (qw/$Revision: 1.1.1.1 $/ )[1] . ")";
my @MY_DEPENDS =
(
  "TIGR::Foundation"
);

# Reference to TF object
my $tf = new TIGR::Foundation;

my $WarnFile = "tmp.warnings";  # data integrity warnings go here

# Operating modes
my $debug = 0;               # Debug level
my $silent = 0;

# Constants
my %NUCLEOTIDES = 
(
  "a"  => 1,
  "c"  => 2,
  "t"  => 4,
  "g"  => 8,
  "m"  => 3,
  "r"  => 9,
  "w"  => 5,
  "s"  => 10,
  "y"  => 6,
  "k"  => 12,
  "v"  => 11,
  "h"  => 7,
  "d"  => 13,
  "b"  => 14,
  "x"  => 15,
  "n"  => 15,
);
my %AMBIGUITIES = 
(
  1  => "a",
  2  => "c",
  4  => "t",
  8  => "g",
  3  => "m",
  9  => "r",
  5  => "w",
  10 => "s",
  6  => "y",
  12 => "k",
  11 => "v",
  7  => "h",
  13 => "d",
  14 => "b",
  15 => "x",
  15 => "n",
);
my %TRANSITIONS = 
(
  "ac" => 1,
  "ag" => 1,
  "ct" => 1,
  "gt" => 1,
  "at" => 0,
  "cg" => 0, 
);
my $FLANK = 0;   # Default Number of bases to quote on either side of feature
my $flank = $FLANK;  # What's actually used.
my $NOBASE = '.';  # What to display when the flank exceeds the sequence

# ================================== Main =====================================
MAIN:
{
  my $prefix = "";                  # project prefix
  my $alignfile = undef;
  my $print_path = 0;               # whether to print full path of the operands
  my $user_chain = undef;           # which chain if any to output only

  # ========================== Program Setup ==============================
  # Prepare logs
  $tf->addDependInfo(@MY_DEPENDS);
  $tf->setHelpInfo($HELPTEXT);
  $tf->setVersionInfo($MY_VERSION);

  # now we handle the input options
  my $result  = $tf->TIGR_GetOptions
                (
                  'flank=i',    \$flank,
                  'path!',      \$print_path,
                  'chain=i',    \$user_chain,
                );

  $tf->bail("Command line parsing failed") if ($result == 0);
  # Update debug level if it's specified by the user
  if (! defined $tf->getDebugLevel())
  {
    $tf->setDebugLevel($debug);
  }
  $debug = $tf->getDebugLevel();

  # Log our invocation
  $tf->logLocal("$0 @ARGV", 4);

  # Get prefix and bind filenames
  $alignfile = $ARGV[0];
  $tf->bail("Must specify a .align file ") if (! defined $alignfile);
  my ($name, $path, $suffix) = fileparse($alignfile, ".align");
  $prefix = $name; 
  my $outprefix = $prefix;
  $WarnFile = "$outprefix.warnings";  # data integrity warnings go here
  unlink $WarnFile;                   # Truncate the .warnings file

  # =========================== Gather Inputs =============================
  my $af = new IO::File("< $alignfile")
    or $tf->bail("Could not open .align file $alignfile ($!)");
  my $lines = 0;
  my $reffile = "";
  my $queryfile = "";
  my $refacc = "";
  my $queryacc = "";
  my $rnucs = "";
  my $qnucs = ""; 
  my $snps  = "";
  my $bref = 0;
  my $eref = 0;
  my $bquery = 0;
  my $equery = 0;
  my $chain = 0;
  my $state = 0;
  while (my $line = $af->getline())
  {
    $lines++;
    chop $line;
    next if ($line eq "");
    next if ($line =~ m/^==/o); 

    if ($lines == 1)
    {
      # Recon first line as the two alignment object filenames
      # Two strings, must be new alignment objects 
      my @files = split /\s+/,$line; 
      my $here = $tf->getProgramInfo('env_path');  # canonical path
      $reffile   = $files[0];
      $queryfile = $files[1];
      if ($print_path)
      {
        $reffile   = ($reffile eq basename($reffile))? "$here/$reffile" : $reffile;
        $queryfile = ($queryfile eq basename($queryfile))? "$here/$queryfile" : $queryfile;
        $reffile   = ($reffile   =~ m/^\.\.\//o)? "$here/$reffile" : $reffile;
        $queryfile = ($queryfile =~ m/^\.\.\//o)?   "$here/$queryfile" : $queryfile;
      }
      else
      {
        $reffile   = basename($reffile);
        $queryfile = basename($queryfile);
      }
    }
    elsif ($line =~ m/-- Alignments between (.+) (.+)/o) 
    {
      $refacc = (split /\s+/,$1)[0];
      $queryacc = $2; 
    }
    # Initialize a new chain
    elsif ($line =~ m/\-\- BEGIN alignment \[ \+1 (\d+) \- (\d+) \| [\+-]1 (\d+) \- (\d+) \]/o) 
    {
      $bref = $1;
      $eref = $2;
      $bquery = $3;
      $equery = $4;
      $rnucs = "";
      $qnucs = ""; 
      $snps  = "";
      $state = 1;
    }
    elsif ($line =~ m/(\d)\s+([actg\.mrwsykvhdbxnACTGMRWSYKVHDBXN]+)/g)
    {
      if ($state == 1)     
      {
        $rnucs .= $2;
        $state = 2;
      } 
      else
      {
        $qnucs .= $2;
        $state = 1;
      }
    }
    elsif ($line =~ m/\^/g)
    {
      $snps .= substr($line, 11);
      $state = 1;
    }
    elsif ($line =~ m/\-\-\s+END alignment/o) 
    {
      if ( (defined $user_chain && $chain == $user_chain) ||
           (!defined $user_chain)
         )
      {
        makeSNPs($reffile, 
                 $refacc, 
                 $queryfile, 
                 $queryacc, 
                 $chain, 
                 $rnucs, 
                 $qnucs, 
                 $snps,  
                 $bref, 
                 $eref, 
                 $bquery, 
                 $equery
                );
      }
      $chain++;   
      $state = 0;
    }
    elsif ((scalar split /\s+/,$line) == 2 )  
    {
      # Two strings, must be new alignment objects 
      my @files = split /\s+/,$line; 
      $reffile = $files[0];
      $queryfile = $files[1];
    }
    else
    {
      logwarning("Ignoring .align input \'$line\'") if ($debug > 0);
    }
  }  
  $af->close();

  # Cleanup 
  exit(0);
}


# ================================= Procedures ================================


# logwarning - Print a data warning or error message to stdout
#       unless run with the -silent option, and record the message
#       to the application warnings file.
#       A copy is written to the log file under debug level 0.
# $_[0] The message
#
sub logwarning($)
{
  my $msg = shift;
  if (! defined $msg)
  {
    return;
  }

  # Report to the terminal
  if (! $silent)
  {
    print STDOUT "$msg\n";
  }

  # Messages (without logging info) go to errors file
  my $fh = new IO::File(">> $WarnFile") 
    or $tf->bail("Failed to open $WarnFile ($!)");
  $fh->print("$msg\n");
  $fh->close();

  # Log it at debug level 4
  $tf->logLocal($msg, 4);
}

# Make SNP records out of the .align extents in a chain
sub makeSNPs($$$$$$$$$$$$)
{
  my ($reffile, 
      $refacc, 
      $queryfile, 
      $queryacc,  
      $chain, 
      $rnuc, 
      $qnuc, 
      $snp,  
      $bref, 
      $eref, 
      $bquery, 
      $equery,
     ) = @_;
  my  @rnucs = split //,$rnuc;
  my  @qnucs = split //,$qnuc;
  my  @lsnps = ();
  my $strand = ($equery < $bquery)? 1 : 0;

  my $k=1;
  my $r=1;
  my $q=1;
  my $len = length($rnuc);  # they are all the same length
  my $i=0;
  my $emit = 0;
  my $bases = ""; 
  my $transition = 0; 
  my $rcoord = 0; 
  my $qcoord = 0; 
  my $type = "";
  my $qextent = ""; 
  my $rextent = ""; 
  my $rgaps = 0;
  my $qgaps = 0;
  my $perfect = 1;

  # Detect whether reference sequence is aligned with leading or trailing gaps
  my $leading_rgaps  = ($rnucs[0]  eq '.')? 1 : 0;
  my $trailing_rgaps = ($rnucs[-1] eq '.')? 1 : 0;
  my $n_leading_rgaps = 0;
  my $k=0;
  while ($rnucs[$k++] eq '.')
  {
    $n_leading_rgaps++;
  }
  # Detect whether query sequence is aligned with leading or trailing gaps
  my $leading_qgaps  = ($qnucs[0]  eq '.')? 1 : 0;
  my $trailing_qgaps = ($qnucs[-1] eq '.')? 1 : 0;
  my $n_trailing_qgaps = 0;
  $k=-1;
  while ($qnucs[$k--] eq '.')
  {
    $n_trailing_qgaps++;
  }

  while ($i < $len) 
  {
    if ($rnucs[$i] eq $qnucs[$i])
    {
      $emit = 0;  # agrees, do nothing
      $i++;
    }
    elsif ($rnucs[$i] ne '.'  &&  $qnucs[$i] ne '.'  &&  $rnucs[$i] ne $qnucs[$i])
    {
      # SNP
      my $rflags = $NUCLEOTIDES{lc($rnucs[$i])};
      my $qflags = $NUCLEOTIDES{lc($qnucs[$i])};
      $bases = $AMBIGUITIES{$rflags | $qflags}; 
      my $pattern = join "", sort ($rnucs[$i],$qnucs[$i]);
      $transition = (exists $TRANSITIONS{$pattern})? $TRANSITIONS{$pattern} : 2;
      $rcoord = $bref + $i - $rgaps; 
      $qcoord = ($strand)?  $bquery - $i + $qgaps: $bquery + $i - $qgaps;
      $type = "S";
      $qextent = lflank($i, $flank, $qnuc) . $qnucs[$i] . rflank($i, $flank, $qnuc);
      $rextent = lflank($i, $flank, $rnuc) . $rnucs[$i] . rflank($i, $flank, $rnuc);
      $emit = 1;
      $perfect = 0;
      $i++;
    } 
    elsif ($rnucs[$i] eq '.'  &&  $qnucs[$i] ne '.')
    {
      # indel, insertion in query sequence
      # read the indel extent
      $bases = ""; 
      my $dels = "";
      $transition = 2;  # neither transition nor transversion

      # Special cases for deletion required to start of reference   
      # (leading alignment gaps in query).   Note that reference is
      # always forward.
      # 
      if ($strand == 0)
      {
        if ($leading_rgaps && $i == 0)
        { 
          # Leading insert from reference in query, forward strand: eg
          # 1          ....aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          # 1          aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          #
          $rcoord = 0;  
          $rgaps = 1;
          $qcoord = ($strand)?  $bquery - $i + $qgaps : $bquery + $i - $qgaps;
        }
        else
        { 
          $rcoord = $bref + $i - $rgaps;
          $qcoord = $bquery + $i - $qgaps;
        }
      }
      else
      {
        if ($leading_rgaps && $i == 0)
        { 
          # Leading insert from reference in query, reverse strand: eg
          # 1          ....aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          # 250        aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          #
          $rcoord = 0;
          $rgaps = 1;
          $qcoord = $bquery - $n_leading_rgaps + 1;  
        }
        else
        { 
          $rcoord = $bref + $i - $rgaps;
          # Need to move the starting point on the query to where the
          # indel starts (lower order coordinate).
          my $j = $i;
          while ($rnucs[$j++] eq '.') { ; }
          $qcoord = $bquery - ($j - 2) + $qgaps;
        }
      }
      
      
      $type = "I";
      my $mark = $i;
      while ($rnucs[$i] eq '.'  &&  $qnucs[$i] ne '.'  &&  $i < $len)
      {
        $bases .= $qnucs[$i];
        $dels  .= $rnucs[$i];
        $rgaps++ if ($mark != $i);
        $i++;
      }
      my $set = $i - 1;
      $qextent = lflank($mark, $flank, $qnuc) . $bases . rflank($set, $flank, $qnuc); 
      $rextent = lflank($mark, $flank, $rnuc) . $dels  . rflank($set, $flank, $rnuc); 

      $emit = 1;
      $perfect = 0;
      # don't increment $i
    }
    elsif ($rnucs[$i] ne '.'  &&  $qnucs[$i] eq '.')
    {
      # indel, deletion from query sequence
      # read the indel extent
      $bases = "";
      my $dels = "";
      $transition = 2;  # neither transition nor transversion
      $rcoord = $bref + $i - $rgaps;

      # Special cases for deletion required to start of reference   
      # (leading alignment gaps in query).  
      # 
      if ($strand == 0)
      {
        if ($leading_qgaps && $i == 0)
        { 
          # Leading delete from reference in query, forward strand: eg
          # 1          aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          # 1          ....aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          #
          $qcoord = 0;  
          $qgaps = 1;
        }
        elsif ($trailing_qgaps && $i >= ($len - $n_trailing_qgaps))
        { 
          # Trailing delete from reference in query, forward strand: eg
          # 1          aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          # 1          aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa....
          #
          $qcoord = $equery;             
        }
        else
        { 
          $qcoord = $bquery + $i - $qgaps;
        }
      }
      else
      {
        if ($leading_qgaps && $i == 0)
        { 
          # Leading delete from reference in query, reverse strand: eg
          # 1          aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          # 250        ....aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          #
          $qcoord = $bquery;  
          $qgaps = 1;
        }
        elsif ($trailing_qgaps && $i >= ($len - $n_trailing_qgaps))
        { 
          # Leading delete from reference in query, reverse strand: eg
          # 99         aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
          # 31         aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa....
          #
          $qcoord = 0;             
        }
        else
        { 
          $qcoord = $bquery - $i + $qgaps;
          $qcoord--;   # refer to the qcoord less than the start of gap 
        }
      }
      
      $type = "D";
      my $mark = $i;
      while ($rnucs[$i] ne '.'  &&  $qnucs[$i] eq '.'  &&  $i < $len)
      {
        $bases .= $rnucs[$i];
        $dels  .= $qnucs[$i];
        $qgaps++ if ($mark != $i);
        $i++;
      }
      my $set = $i - 1;
      $qextent = lflank($mark, $flank, $qnuc) . $dels  . rflank($set, $flank, $qnuc); 
      $rextent = lflank($mark, $flank, $rnuc) . $bases . rflank($set, $flank, $rnuc); 
      $emit = 1;
      $perfect = 0;
      # don't increment $i
    }
    if ($perfect  &&  $i == $len)
    {
      # No polymorphisms detected due to perfect alignment
      $bases = ".";
      $transition = 2;  # neither transition nor transversion
      $rcoord = 0; 
      $qcoord = 0;
      $type = "N";
      $qextent = '.';
      $rextent = '.'; 
      $i++;
      $emit = 1;
    }
    if ($emit)
    {
      my $strandS = ($strand)? "R" : "F";
      print "$reffile, $refacc, $queryfile, $queryacc, " . 
            "$chain, $bref, $eref, $bquery, $equery, $strandS, " .
            "$rcoord, $qcoord, $bases, $transition, $type, $rextent, $qextent\n";
    } 
    $rgaps++ if ($rnucs[$i] eq '.');
    $qgaps++ if ($qnucs[$i] eq '.');
  }
}

# Compute the left flank of the SNP or indel motif
sub lflank($$$)
{
  my ($i, $flank, $nuc) = @_;

  my $flanklen = ($i < $flank)? $i : $flank; 
  my $empty = "";
  for (my $j=0; $j < $flank - $flanklen; $j++)
  {
    $empty .= $NOBASE;
  }
  my $flanks = $empty . substr($nuc, $i - $flanklen, $flanklen); 
  return ($flanks);
}

# Compute the right flank of the SNP or indel motif
sub rflank($$$)
{
  my ($i, $flank, $nuc) = @_;
  my $len = length($nuc);

  my $flanklen = ($i + 1 > $len - $flank)? $len - $i - 1: $flank; 
  my $empty = "";
  for (my $j=0; $j < $flank - $flanklen; $j++)
  {
    $empty .= $NOBASE;
  }
  my $flanks = substr($nuc, $i + 1, $flanklen) . $empty; 
  return ($flanks);
}

