#include "ContigAlign.hh"

ContigAlign::ContigAlign(bool verbose)
{
  m_verbose = verbose;

}

ContigAlign::~ContigAlign()
{

}

void ContigAlign::initializePublicAPI(Contig * contig)
{
  m_contig = contig;
  m_gindex = m_contig->getWidth();
}


long int ContigAlign::getBasesToSkip(DeltaVector & delta,
                                     DeltaVector::const_iterator & deltaiter,
                                     bool & consGap,
                                     unsigned int cpos,
                                     unsigned int rpos,
                                     const string & trimmed,
                                     const string & consensus,
                                     bool autoPromoteConsGaps)
{
  long int basesToSkip = 0;

  if (deltaiter != delta.end())
  {
    basesToSkip = *deltaiter;
    consGap = true;

    if (basesToSkip < 0)
    {
      basesToSkip = -basesToSkip;
      consGap = false;
    }

    if (!autoPromoteConsGaps) { consGap = !consGap; }

    if (m_verbose)
    {
      if (cpos < (unsigned int) m_gindex && 
          rpos < trimmed.size())
      {
        cerr << "basesToSkip: " << basesToSkip 
             << " cpos: " << cpos << " [" << consensus[cpos] << "]"
             << " rpos: " << rpos << " [" << trimmed[rpos]     << "]"
             << " consGap: " << consGap
             << endl;
      }
    }
  }

  return basesToSkip;
}



// Merge the bases in trimmedSeq with the consensus beteen cpos and lastConsPos
// (inclusively) inserting gaps as necessary. Throw an error if the alignment
// does go as expected. Optionally autoPromote gaps from the consensus as
// necessary. lastConsPos must point at an actual base (not gap).

void ContigAlign::mergeExtension(unsigned int & cpos, 
                                 unsigned int & lastConsPos,
                                 unsigned int & rpos, 
                                 string & trimmedSeq, 
                                 string & consensus,
                                 ContigSequence * read,
                                 DeltaVector & delta,
                                 bool autoPromoteConsGaps,
                                 int consMovementDelta)
{
  int startcpos = cpos;

  if (m_verbose)
  {
    cerr << "Delta:"; 
    for (unsigned int j = 0; j < delta.size(); j++) { cerr << " " << delta[j]; }
    cerr << endl;
  }

  bool consGap = true;
  DeltaVector::const_iterator deltaiter = delta.begin();
  long int basesToSkip = getBasesToSkip(delta, deltaiter, consGap,
                                        cpos, rpos, trimmedSeq, consensus,
                                        autoPromoteConsGaps);

  // AutoPromote Consensus Gaps into the trimmedSeq, and
  // merge the delta values from the alignment
  while(cpos >= 0 && cpos < (unsigned int) m_gindex &&
        cpos <= lastConsPos && rpos < trimmedSeq.size())
  {

    // cerr <<  "cons[" << cpos << "]=" << consensus[cpos]
    //      << " read[" << rpos << "]=" << trimmedSeq[rpos] << endl;

    if (basesToSkip == 1)
    {
      if (consGap)
      {
        if (autoPromoteConsGaps && consensus[cpos] == '-')
        {
          // next position is a gap anyways
          if (m_verbose) { cerr << "Gap already present at " << cpos << endl; }
        }
        else
        {
          // insert a gap
          m_gindex = m_contig->insertGapSlice(cpos, read);
          lastConsPos++;
        }
      }
      else
      {
        if (m_verbose) { cerr << "inserting gap into r at " << cpos  << endl; }
        trimmedSeq.insert(rpos, "-");

        while (autoPromoteConsGaps && consensus[cpos] == '-')
        {
          if (m_verbose) { cerr << "promoting adjacent consensus gap at " << cpos << endl; }
          trimmedSeq.insert(rpos, "-");
          rpos++; cpos += consMovementDelta;
        }
      }

      deltaiter++;
      basesToSkip = getBasesToSkip(delta, deltaiter, consGap, 
                                   cpos+consMovementDelta, 
                                   rpos+1, trimmedSeq, consensus,
                                   autoPromoteConsGaps);
    }
    else
    {
      if (autoPromoteConsGaps && consensus[cpos] == '-')
      {
        if (m_verbose) { cerr << "promoting consensus gap at " << cpos << endl; }
        trimmedSeq.insert(rpos, "-");
      }
      else if (basesToSkip)
      {
        // Don't count gaps
        basesToSkip--;
      }
    }
    
    // Step to next position
    cpos += consMovementDelta; 
    rpos++;
  }

  // handle special case where gaps need to be inserted at end of alignment, 
  // after the read or consensus has "ended".
  if (basesToSkip)
  {
    if (m_verbose) { cerr << "Extra Bases to skip: " << basesToSkip << endl; }

    while (basesToSkip == 1)
    {
      if (consGap)
      {
        // Don't use gaps that are already present since we hit the end of allow range
        m_gindex = m_contig->insertGapSlice(cpos, read);
        lastConsPos++;
      }
      else
      {
        while (autoPromoteConsGaps && consensus[cpos] == '-')
        {
          if (m_verbose) { cerr << "promoting consensus gap at " << cpos << endl; }
          trimmedSeq.insert(rpos, "-");

          rpos++;
          cpos += consMovementDelta;
        }

        if (m_verbose) { cerr << "Appending extra gap to fix alignment!" << endl; }
        trimmedSeq.insert(rpos, "-");
      }

      deltaiter++;
      basesToSkip = getBasesToSkip(delta, deltaiter, consGap, 
                                   cpos+consMovementDelta, 
                                   rpos+1, trimmedSeq, consensus,
                                   autoPromoteConsGaps);
      rpos++;
      cpos += consMovementDelta;
    }
  }

  if (m_verbose)
  {
    cerr << "final:"
         << " cons[" << cpos-1 << "]=" << consensus[cpos-1]
         << " read[" << rpos-1 << "]=" << trimmedSeq[rpos-1]
         << " read[" << trimmedSeq.length()-1 << "]=" << trimmedSeq[trimmedSeq.length()-1]
         << endl;
  }

  if (basesToSkip)
  {
    cerr << "Internal Error!!! More delta values that sequence to align!" << endl;
    throw Exception("More delta values than sequence to align!",
                    "Contig::mergeExtension");
  }

  if (rpos != trimmedSeq.length() || cpos != lastConsPos+1)
  {
    char buffer[1024];
    sprintf(buffer, "Attempt to merge alignment failed: rpos=%d rlen=%d cpos=%d lastcpos=%d",
            rpos, trimmedSeq.length(), cpos, lastConsPos);
    throw Exception(buffer, "Contig::mergeExtension");
  }

  // just mark the tiling dirty
  m_contig->addReadToTiling(read, startcpos, trimmedSeq.length());
}
