#include "RequestManager.hh"

RequestManager::RequestManager(const string & logCategory)
  : m_out(m_outfileStream ? (ostream &)*m_outfileStream : cout),
    m_log(m_tf->getLogger()->getCategory(logCategory))
{
  m_requestData = NULL;
  m_readManager = NULL;

  m_log.setMessageLevel(MESSAGE_LEVEL_DEBUG_HIGH);
  m_doLogging = m_tf->getOptions()->hasOption("debug") &&
                 (m_log.getMessageLevel() <= m_log.getFilterLevel());

  // Should probably always be set to false
  libSlice_setRecallEmpty(m_recallEmpty); 

  registerOptionHandler("HighQualityThreshold", setHighQualityThreshold);
  registerOptionHandler("AmbiguityMethod",      setDoAmbiguity);
  registerOptionHandler("DoConsensus",          setDoConsensus);
  registerOptionHandler("ScoringModel",         setScoringModel);
}

// Static Members
int               RequestManager::m_recallEmpty(0);
int               RequestManager::m_highQualityThreshold(0);
int               RequestManager::m_doConsensus(0);
int               RequestManager::m_doAmbiguity(0);
int               RequestManager::m_errorCount(0);
int               RequestManager::m_scoringModel(2);
int               RequestManager::m_doStats(0);
string            RequestManager::m_outfile("");
ofstream *        RequestManager::m_outfileStream(NULL);
Foundation * RequestManager::m_tf(NULL);

RequestManager::~RequestManager()
{ 
  if (m_requestData) delete m_requestData;

  if (m_outfileStream) 
  {
    // This must be set to NULL since it is a static variable
    delete m_outfileStream;
    m_outfileStream = NULL;
  }
}

void RequestManager::addReadManager(ReadManager * readManager)
{
  m_readManager = readManager;
}

void RequestManager::addReadData(ReadData * readData)
{ }

void RequestManager::setDoStats(int doStats)
{
  m_doStats = doStats;
}

void RequestManager::setHighQualityThreshold(int highQualityThreshold)
{
  m_highQualityThreshold = highQualityThreshold;
}

void RequestManager::setDoAmbiguity(int doAmbiguity)
{
  m_doAmbiguity = doAmbiguity;
}

void RequestManager::setDoConsensus(int doConsensus)
{
  m_doConsensus = doConsensus;
}

void RequestManager::setScoringModel(int scoringModel)
{
  m_scoringModel = scoringModel;
}

void RequestManager::setDoAmbiguity(const string & value)
{
  if (value == "Conic" || value == "3")
  {
    m_doAmbiguity = 3;
  }
  else if (value == "Annotation" || value == "2") 
  {
    m_doAmbiguity = 2;
  }
  else if (value == "Minimal" || value == "1")
  {
    m_doAmbiguity = 1;
  }
  else
  {
    m_doAmbiguity = 0;
  }
}

void RequestManager::setScoringModel(const string & value)
{
  m_scoringModel = atoi(value.c_str());
}

void RequestManager::setDoConsensus(const string & value)
{
  m_doConsensus = (value == "0") ? 0 : 1;
}

void RequestManager::setHighQualityThreshold(const string & value)
{
  m_highQualityThreshold = atoi(value.c_str());
}

void RequestManager::addSimpleOption(const char * name,
                                     const char * value)
{
  string strvalue = (value) ? value : "";

  OptionMap::iterator i = m_optionMap.find(name);

  if (i != m_optionMap.end())
  {
    OptionHandler callback = i->second;
    (*callback)(strvalue);
  }
  else
  {
    throw Exception((string) "Option \"" + name + "\" not understood",
                    "RequestManager::addSimpleOption");
  }
}

void RequestManager::registerOptionHandler(const string & name,
                                           OptionHandler callback)
{
  m_optionMap[name] = callback;
}

void RequestManager::addRequest(ConsensusRequest * requestData)
{
  if (m_requestData)
  {
    delete m_requestData;
  }

  m_requestData = requestData;
}

void RequestManager::addSliceRange(long & index, long & gindex)
{ 
  index = gindex = 0L;
}

void RequestManager::addSlice(SliceData * slice)
{ 
  slice->setHighQualityThreshold(m_highQualityThreshold);
  slice->setDoAmbiguity(m_doAmbiguity);
  slice->setScoringModel(m_scoringModel);

  if (m_doLogging)   { m_log << slice; }

  if (m_doConsensus) 
  {
    if (!slice->getNucData().empty() || m_recallEmpty)
    {
      // Recall the consensus if there are bases or requested
      slice->doCalculations(); 
    }
  }
}

int RequestManager::printResults()
{ 
  return m_errorCount;
}

void RequestManager::printStats(ostream & out, const string & prefix)
{ }

void RequestManager::errorSliceHandler(SliceData * slice, 
                                       const Exception & e)
{
  m_errorCount++;
                       slice->printError(m_out);
  if (m_out != cout) { slice->printError(cerr); }

  // Also log the error
  m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_ERROR);
  //slice->printError((ostream &) m_log);
  
  m_log << "Internal error in slice: " << slice->getGindex() << endl
        << "In file: " << m_infile << endl;

  m_log.popMessageLevel();

  delete slice;
}


void RequestManager::printXMLHeader(ConsensusRequest * requestData,
                                    ostream * outstream)
{
  if (outstream)
  {
    *outstream << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
    *outstream << "<!DOCTYPE Request SYSTEM \"" << SLICE_DTD << "\">" << endl;

    if (requestData)
    {
      requestData->printXML(*outstream);
    }

  }
  else
  {
    m_out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << endl;
    m_out << "<!DOCTYPE Request SYSTEM \"" << SLICE_DTD << "\">" << endl;

    if (requestData)
    {
      requestData->printXML(m_out);
    }
  }
}


ofstream * RequestManager::openOutputFile(const string & filename)
{
  std::ofstream * retval = NULL;

  if (filename.length())
  {
    retval = new std::ofstream (filename.c_str());

    if (!retval || !retval->is_open())
    {
      throw UnableToOpenError(Exception::quote(filename),
                              "RequestManager::openOutputFile");
    }
  }

  return retval;
}




//! Parses and gets results for a single file
/*! Once the parser has been initialized, this function is called to actually
 *  process the XML file, and print the results.
 */
int RequestManager::processFile(const string & infile, ostream & error)
{
  int retval = 0;
  m_infile = infile;

  try
  {
    XMLPlatformUtils::Initialize();
  }
  catch (const XMLException & e)
  {
    m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_ERROR);
    m_log << "Error during XML initialization! Message: " 
          << StrX(e.getMessage()) << endl;
    m_log.popMessageLevel();

    error << "Error during XML initialization! Message: " 
          << StrX(e.getMessage()) << endl;

    return 10;
  }

  SAX2XMLReader* parser = NULL;

  try
  {
    // Initalize the Parser
    parser = XMLReaderFactory::createXMLReader();
    parser->setFeature(XMLUni::fgSAX2CoreValidation, true);
    parser->setFeature(XMLUni::fgXercesDynamic,      true);

    LogCategory saxlog = m_tf->getLogger()->getCategory("SAX");
    SliceSAX2Handler handler(this, saxlog,
                             m_tf->getOptions()->hasOption("debug"));

    parser->setContentHandler(&handler);
    parser->setErrorHandler(&handler);

    if (infile == "-")
    {
      // Process data from stdin
      StdInInputSource in;
      parser->parse(in);
    }
    else
    {
      // Process data loaded from a filename
      parser->parse(infile.c_str());
    }

    retval = printResults();

    if (m_doStats) { handler.printStats(cerr, infile); }
  }
  catch (const SAXParseException & e)
  {
    m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_ERROR);
    m_log << "Error at file=" << StrX(e.getSystemId())
          << " line="         << e.getLineNumber()
          << " char="         << e.getColumnNumber()
          << " Message="      << StrX(e.getMessage())
          << endl;
    m_log.popMessageLevel();

    error << "Error at file=" << StrX(e.getSystemId())
          << " line="         << e.getLineNumber()
          << " char="         << e.getColumnNumber()
          << " Message="      << StrX(e.getMessage())
          << endl;

    retval = 101;
  }
  catch (const XMLException & e)
  {
    m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_ERROR);
    m_log << "Error during processing: '" << infile << "'"
          << "Message is: " << StrX(e.getMessage()) 
          << endl;
    m_log.popMessageLevel();

    error << "Error during processing: '" << infile << "'"
          << "Message is: " << StrX(e.getMessage()) 
          << endl;

    retval = 102;
  }
  catch (const Exception & e)
  {
    m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_ERROR);
    m_log << "Error during processing: '" << infile << "'" << endl
          << e << endl;
    m_log.popMessageLevel();

    error << "Error during processing: '" << infile << "'" << endl
          << e << endl;

    retval = 103;
  }
  catch (const SafeException & e)
  {
    // Needed to break the processing for a known reason
  }
  catch (...)
  {
    m_log.pushMessageLevel(MESSAGE_LEVEL_DEBUG_ERROR);
    m_log << "Unexpected exception during processing: '" 
          << infile << "'" << endl;
    m_log.popMessageLevel();

    error << "Unexpected exception during processing: '" 
          << infile << "'" << endl;

    retval = 104;
  }

  if (parser) delete parser;
  XMLPlatformUtils::Terminate();

  return retval;
}

void RequestManager::addStandardOptions(Foundation * tf)
{
  m_tf = tf;
  tf->getOptions()->addOptionResult("output|o=s", &m_outfile,
                                    "Specify the output file");

  tf->getOptions()->addOptionResult("consensus|c", &m_doConsensus,
                                    "Calculate the consensus");

  tf->getOptions()->addOptionResult("ambiguity|a=i", &m_doAmbiguity,
                                    "0=No ambiguity, 1=Churchill&Waterman, 2=Annotation, 3=Conic");

  tf->getOptions()->addOptionResult("scoring|m=i", &m_scoringModel,
                                    "Scoring model for conic ambiguities");

  tf->getOptions()->addOptionResult("highQuality|t=i", &m_highQualityThreshold,
                                    "Set the High Quality Threshold");

  tf->getOptions()->addOptionResult("stats|s", &m_doStats,
                                    "Display Statistics");

  tf->getOptions()->addOptionResult("recall-empty|e", &m_recallEmpty,
                                    "Recall the consensus of empty slices");
}


void RequestManager::handleOptions(Foundation * tf)
{
  addStandardOptions(tf);
  tf->handleStandardOptions();
  handleOptionResult();
}

void RequestManager::handleOptionResult()
{
  // Setup the output stream
  if (m_outfile.length())
  {
    m_outfileStream = RequestManager::openOutputFile(m_outfile);
  }
}

string RequestManager::getOptionHelp()
{
  return
"\n"
"   Slice Options\n"
"   -------------\n"
"   -s|--stats  Display Statistics about processing\n"
"   -o|--output Specify output file (default to stdout)\n"
"\n"
"   -c|--consensus    Recall the consensus of each slice\n"
"   -e|--recall-empty Recall the consensus of empty slices (Containing no bases)\n"
"   -t|--highQuality <threshold> Specify the high quality threshold\n"
"\n"
"   -a|--ambiguity <mode> Use the specified mode of ambiguity calling\n"
"      0 = No Ambiguity Codes\n"
"      1 = Churchill and Waterman Ambiguity Codes\n"
"      2 = Annotation Ambiguity Codes\n"
"      3 = Conic Ambiguity Codes\n"
"\n"
"Please see http://intranet/software_docs/SliceTools.html for more information.\n";
}

string RequestManager::getVersionString(char * gVersionStr,
                                        char * gRevisionStr)
{
  gRevisionStr[strlen(gRevisionStr)-2] = '\0';
  string version = (string) gVersionStr +
                   (string) " (Build" + (string)(gRevisionStr+9) +
                   (string) ")\n  (" + libSlice_getVersion() +  ")";

  return version;
}

ostream & RequestManager::getOutputStream() const
{
  return m_out;
}

void RequestManager::print60(ostream & out, const string & bases)
{
  int count = 0;
  int printNewline = 1;

  string::const_iterator i;

  for (i = bases.begin();
       i != bases.end();
       i++)
  {
    out << *i;
    count++;
    printNewline = 1;

    if (count == 60)
    {
      out << endl;
      count = 0;

      printNewline = 0;
    }
  }

  if (printNewline) { out << endl; }
}

